@@ -1077,21 +1077,21 @@ prolog_message(query(QueryResult)) -->
 query_result(no) -->            % failure
     [ ansi([bold,fg(red)], 'false.', []) ],
     extra_line.
-query_result(yes([])) -->      % prompt_alternatives_on: groundness
+query_result(yes(true, [])) -->      % prompt_alternatives_on: groundness
     !,
     [ ansi(bold, 'true.', []) ],
     extra_line.
-query_result(yes(Residuals)) -->
-    result([], Residuals),
+query_result(yes(Delays, Residuals)) -->
+    result([], Delays, Residuals),
     extra_line.
 query_result(done) -->          % user typed <CR>
     extra_line.
-query_result(yes(Bindings, Residuals)) -->
-    result(Bindings, Residuals),
-    prompt(yes, Bindings, Residuals).
-query_result(more(Bindings, Residuals)) -->
-    result(Bindings, Residuals),
-    prompt(more, Bindings, Residuals).
+query_result(yes(Bindings, Delays, Residuals)) -->
+    result(Bindings, Delays, Residuals),
+    prompt(yes, Bindings, Delays, Residuals).
+query_result(more(Bindings, Delays, Residuals)) -->
+    result(Bindings, Delays, Residuals),
+    prompt(more, Bindings, Delays, Residuals).
 query_result(help) -->
     [ nl, 'Actions:'-[], nl, nl,
       '; (n, r, space, TAB): redo    t:          trace & redo'-[], nl,
@@ -1109,10 +1109,10 @@ query_result(eof) -->
 query_result(toplevel_open_line) -->
     [].
 
-prompt(Answer, [], []-[]) -->
+prompt(Answer, [], true, []-[]) -->
     !,
     prompt(Answer, empty).
-prompt(Answer, _, _) -->
+prompt(Answer, _, _, _) -->
     !,
     prompt(Answer, non_empty).
 
@@ -1131,13 +1131,24 @@ prompt(more, _) -->
     !,
     [ ' '-[], flush ].
 
-result(Bindings, Residuals) -->
+result(Bindings, Delays, Residuals) -->
     { current_prolog_flag(answer_write_options, Options0),
-      Options = [partial(true)|Options0]
+      Options = [partial(true)|Options0],
+      GOptions = [priority(999)|Options0]
     },
+    wfs_residual_program(Delays, GOptions),
     bindings(Bindings, [priority(699)|Options]),
-    bind_res_sep(Bindings, Residuals),
-    residuals(Residuals, [priority(999)|Options]).
+    (   {Residuals == []-[]}
+    ->  bind_delays_sep(Bindings, Delays),
+        delays(Delays, GOptions)
+    ;   bind_res_sep(Bindings, Residuals),
+        residuals(Residuals, GOptions),
+        (   {Delays == true}
+        ->  []
+        ;   [','-[], nl],
+            delays(Delays, GOptions)
+        )
+    ).
 
 bindings([], _) -->
     [].
@@ -1197,11 +1208,49 @@ residuals1([G|Gs], Options) -->
     ;   [ '~W'-[G, Options] ]
     ).
 
+wfs_residual_program(true, _Options) -->
+    !.
+wfs_residual_program(Goal, _Options) -->
+    { current_prolog_flag(toplevel_list_wfs_residual_program, true),
+      !,
+      '$current_typein_module'(TypeIn),
+      (   current_predicate(delays_residual_program/2)
+      ->  true
+      ;   use_module(library(wfs), [delays_residual_program/2])
+      ),
+      delays_residual_program(TypeIn:Goal, TypeIn:Program)
+    },
+    !,
+    [ ansi(fg(green), '% WFS residual program', []), nl ],
+    [ ansi(fg(cyan), '~@', ['$messages':list_clauses(Program)]) ].
+wfs_residual_program(_, _) --> [].
+
+delays(true, _Options) -->
+    !.
+delays(Goal, Options) -->
+    { current_prolog_flag(toplevel_list_wfs_residual_program, true)
+    },
+    !,
+    [ ansi([bold], '~W', [Goal, Options]) ].
+delays(_, _Options) -->
+    [ ansi([bold,fg(cyan)], unknown, []) ].
+
+:- public list_clauses/1.
+
+list_clauses([]).
+list_clauses([H|T]) :-
+    portray_clause(user_output, H, [indent(4)]),
+    list_clauses(T).
+
 bind_res_sep(_, []) --> !.
 bind_res_sep(_, []-[]) --> !.
 bind_res_sep([], _) --> !.
 bind_res_sep(_, _) --> [','-[], nl].
 
+bind_delays_sep([], _) --> !.
+bind_delays_sep(_, true) --> !.
+bind_delays_sep(_, _) --> [','-[], nl].
+
 extra_line -->
     { current_prolog_flag(toplevel_extra_white_line, true) },
     !,
