@@ -937,6 +937,9 @@ define_or_generate(Pred) :-
     '$get_predicate_attribute'(Pred, quasi_quotation_syntax, 1).
 '$predicate_property'(defined, Pred) :-
     '$get_predicate_attribute'(Pred, defined, 1).
+'$predicate_property'(tabled(default), M:Pred) :-
+    '$c_current_predicate'(_, M:'$tabled'(_)),
+    M:'$tabled'(Pred).
 
 system_undefined(user:prolog_trace_interception/4).
 system_undefined(user:prolog_exception_hook/4).
