@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org/projects/xpce/
-    Copyright (c)  2011-2018, University of Amsterdam
+    Copyright (c)  2011-2019, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -320,7 +321,10 @@ process_directive(module(_Name, Export), M, Src) :-
            safe_push_op(P,A,M:N, Src)).
 process_directive(use_module(Spec), _, Src) :-
     !,
-    catch(process_use_module(Spec, Src), _, true).
+    catch(process_use_module1(Spec, Src), _, true).
+process_directive(use_module(Spec, Imports), _, Src) :-
+    !,
+    catch(process_use_module2(Spec, Imports, Src), _, true).
 process_directive(Directive, _, Src) :-
     prolog_source:expand((:-Directive), Src, _).
 
@@ -329,16 +333,16 @@ syntax_flag(var_prefix).
 syntax_flag(allow_variable_name_as_functor).
 syntax_flag(allow_dot_in_atom).
 
-%!  process_use_module(+Imports, +Src)
+%!  process_use_module1(+Imports, +Src)
 %
 %   Get the exported operators from the referenced files.
 
-process_use_module([], _) :- !.
-process_use_module([H|T], Src) :-
+process_use_module1([], _) :- !.
+process_use_module1([H|T], Src) :-
     !,
-    process_use_module(H, Src),
-    process_use_module(T, Src).
-process_use_module(File, Src) :-
+    process_use_module1(H, Src),
+    process_use_module1(T, Src).
+process_use_module1(File, Src) :-
     (   xref_public_list(File, Src,
                          [ exports(Exports),
                            silent(true),
@@ -355,6 +359,24 @@ process_use_module(File, Src) :-
     ;   true
     ).
 
+process_use_module2(File, Imports, Src) :-
+    (   xref_public_list(File, Src,
+                         [ exports(Exports),
+                           silent(true),
+                           path(Path)
+                         ])
+    ->  forall(( member(op(P,T,N), Exports),
+                 member(op(P,T,N), Imports)),
+               safe_push_op(P,T,N,Src)),
+        colour_state_module(Src, SM),
+        (   member(Syntax/4, Exports),
+            member(Syntax/4, Imports),
+            load_quasi_quotation_syntax(SM:Path, Syntax),
+            fail
+        ;   true
+        )
+    ;   true
+    ).
 
 %!  prolog_colourise_query(+Query:string, +SourceId, :ColourItem)
 %
@@ -946,7 +968,7 @@ colourise_goal(Goal, Origin, TB, parentheses_term_position(PO,PC,Pos)) :-
     colour_item(parentheses, TB, PO-PC),
     colourise_goal(Goal, Origin, TB, Pos).
 colourise_goal(Goal, _, TB, Pos) :-
-    Pos = list_position(F,T,Elms,_),
+    Pos = list_position(F,T,Elms,TailPos),
     Goal = [_|_],
     !,
     FT is F + 1,
@@ -954,7 +976,7 @@ colourise_goal(Goal, _, TB, Pos) :-
     colour_item(goal_term(built_in, Goal), TB, Pos),
     colour_item(goal(built_in, Goal), TB, F-FT),
     colour_item(goal(built_in, Goal), TB, AT-T),
-    colourise_file_list(Goal, TB, Elms, any).
+    colourise_file_list(Goal, TB, Elms, TailPos, any).
 colourise_goal(Goal, Origin, TB, Pos) :-
     Pos = list_position(F,T,Elms,Tail),
     callable(Goal),
@@ -1091,10 +1113,12 @@ colourise_meta_args(N, Goal, Module, MetaArgs, TB, [P0|PT]) :-
     colourise_meta_args(NN, Goal, Module, MetaArgs, TB, PT).
 
 colourise_meta_arg(MetaSpec, Arg, TB, Pos) :-
+    nonvar(Arg),
     expand_meta(MetaSpec, Arg, Expanded),
     !,
     colourise_goal(Expanded, [], TB, Pos). % TBD: recursion
 colourise_meta_arg(MetaSpec, Arg, TB, Pos) :-
+    nonvar(Arg),
     MetaSpec == //,
     !,
     colourise_dcg_goals(Arg, //, TB, Pos).
@@ -1145,12 +1169,15 @@ expand_meta(MetaSpec, M:Goal, M:Expanded) :-
     expand_meta(MetaSpec, Goal, Expanded).
 expand_meta(MetaSpec, Goal, Expanded) :-
     integer(MetaSpec),
-    callable(Goal),
-    !,
-    length(Extra, MetaSpec),
-    Goal =.. List0,
-    append(List0, Extra, List),
-    Expanded =.. List.
+    MetaSpec > 0,
+    (   atom(Goal)
+    ->  functor(Expanded, Goal, MetaSpec)
+    ;   compound(Goal)
+    ->  compound_name_arguments(Goal, Name, Args0),
+        length(Extra, MetaSpec),
+        append(Args0, Extra, Args),
+        compound_name_arguments(Expanded, Name, Args)
+    ).
 
 %!  colourise_setof(+Term, +TB, +Pos)
 %
@@ -1279,10 +1306,10 @@ colour_option_values([V0|TV], [T0|TT], TB, [P0|TP]) :-
 %
 %   @param Why is one of =any= or =imported=
 
-colourise_files(List, TB, list_position(F,T,Elms,_), Why) :-
+colourise_files(List, TB, list_position(F,T,Elms,TailPos), Why) :-
     !,
     colour_item(list, TB, F-T),
-    colourise_file_list(List, TB, Elms, Why).
+    colourise_file_list(List, TB, Elms, TailPos, Why).
 colourise_files(M:Spec, TB, term_position(_,_,_,_,[MP,SP]), Why) :-
     !,
     colourise_module(M, TB, MP),
@@ -1306,10 +1333,17 @@ colourise_files(Spec0, TB, Pos, Why) :-
     ;   colour_item(nofile, TB, Pos)
     ).
 
-colourise_file_list([], _, _, _).
-colourise_file_list([H|T], TB, [PH|PT], Why) :-
+%!  colourise_file_list(+Files, +TB, +ElmPos, +TailPos, +Why)
+
+colourise_file_list([], _, [], none, _).
+colourise_file_list(Last, TB, [], TailPos, _Why) :-
+    (   var(Last)
+    ->  colourise_term(Last, TB, TailPos)
+    ;   colour_item(type_error(list), TB, TailPos)
+    ).
+colourise_file_list([H|T], TB, [PH|PT], TailPos, Why) :-
     colourise_files(H, TB, PH, Why),
-    colourise_file_list(T, TB, PT, Why).
+    colourise_file_list(T, TB, PT, TailPos, Why).
 
 resolves_anything(TB, Path) :-
     colour_state_source_id(TB, SourceId),
@@ -1524,7 +1558,8 @@ colourise_dict_kv(Dict, TB, [key_value_position(_F,_T,SF,ST,K,KP,VP)|KV]) :-
 %   Colourise the module export-list (or any other list holding
 %   terms of the form Name/Arity referring to predicates).
 
-colourise_exports([], _, _) :- !.
+colourise_exports([], TB, Pos) :- !,
+    colourise_term_arg([], TB, Pos).
 colourise_exports(List, TB, list_position(F,T,ElmPos,Tail)) :-
     !,
     colour_item(list, TB, F-T),
@@ -1683,17 +1718,19 @@ colourise_meta_declaration(M:Head, Extra, TB,
                                          [ MP,
                                            term_position(_,_,FF,FT,ArgPos)
                                          ])) :-
+    compound(Head),
     !,
     colourise_module(M, TB, MP),
     colour_item(functor, TB, QF-QT),
     colour_item(goal(extern(M),Head), TB, FF-FT),
-    Head =.. [_|Args],
+    compound_name_arguments(Head, _, Args),
     colourise_meta_decls(Args, Extra, TB, ArgPos).
 colourise_meta_declaration(Head, Extra, TB, term_position(_,_,FF,FT,ArgPos)) :-
+    compound(Head),
     !,
     goal_classification(TB, Head, [], Class),
     colour_item(goal(Class, Head), TB, FF-FT),
-    Head =.. [_|Args],
+    compound_name_arguments(Head, _, Args),
     colourise_meta_decls(Args, Extra, TB, ArgPos).
 colourise_meta_declaration([H|T], Extra, TB, list_position(LF,LT,[HP],TP)) :-
     !,
@@ -2024,6 +2061,7 @@ goal_colours(discontiguous(_),       built_in-[predicates]).
 goal_colours(multifile(_),           built_in-[predicates]).
 goal_colours(volatile(_),            built_in-[predicates]).
 goal_colours(public(_),              built_in-[predicates]).
+goal_colours(table(_),               built_in-[predicates]).
 goal_colours(meta_predicate(_),      built_in-[meta_declarations]).
 goal_colours(consult(_),             built_in-[file]).
 goal_colours(include(_),             built_in-[file]).
