@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org/projects/xpce/
-    Copyright (c)  2006-2017, University of Amsterdam
+    Copyright (c)  2006-2019, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -38,6 +39,7 @@
             xref_source/2,              % +Source, +Options
             xref_called/3,              % ?Source, ?Callable, ?By
             xref_called/4,              % ?Source, ?Callable, ?By, ?Cond
+            xref_called/5,              % ?Source, ?Callable, ?By, ?Cond, ?Line
             xref_defined/3,             % ?Source. ?Callable, -How
             xref_definition_line/2,     % +How, -Line
             xref_exported/2,            % ?Source, ?Callable
@@ -80,6 +82,8 @@
 :- use_module(library(pldoc), []).      % Must be loaded before doc_process
 :- use_module(library(pldoc/doc_process)).
 :- endif.
+:- use_module(library(solution_sequences)).
+:- use_module(library(modules)).
 
 :- predicate_options(xref_source/2, 2,
                      [ silent(boolean),
@@ -91,7 +95,7 @@
 
 
 :- dynamic
-    called/4,                       % Head, Src, From, Cond
+    called/5,                       % Head, Src, From, Cond, Line
     (dynamic)/3,                    % Head, Src, Line
     (thread_local)/3,               % Head, Src, Line
     (multifile)/3,                  % Head, Src, Line
@@ -475,7 +479,7 @@ xref_set_prolog_flag(_, _, _, _).
 
 xref_clean(Source) :-
     prolog_canonical_source(Source, Src),
-    retractall(called(_, Src, _Origin, _Cond)),
+    retractall(called(_, Src, _Origin, _Cond, _Line)),
     retractall(dynamic(_, Src, Line)),
     retractall(multifile(_, Src, Line)),
     retractall(public(_, Src, Line)),
@@ -523,17 +527,33 @@ xref_done(Source, Time) :-
 
 %!  xref_called(?Source, ?Called, ?By) is nondet.
 %!  xref_called(?Source, ?Called, ?By, ?Cond) is nondet.
+%!  xref_called(?Source, ?Called, ?By, ?Cond, ?Line) is nondet.
 %
-%   Enumerate the predicate-call relations. Predicate called by
-%   directives have a By '<directive>'.
+%   True  when  By  is  called  from    Called   in  Source.  Note  that
+%   xref_called/3  and  xref_called/4  use  distinct/2  to  return  only
+%   distinct `Called-By` pairs. The  xref_called/5   version  may return
+%   duplicate `Called-By` if Called is called   from multiple clauses in
+%   By, but at most one call per clause.
+%
+%   @arg By is a head term or one of the reserved terms
+%   `'<directive>'(Line)` or `'<public>'(Line)`, indicating the call
+%   is from an (often initialization/1) directive or there is a public/1
+%   directive that claims the predicate is called from in some
+%   untractable way.
+%   @arg Cond is the (accumulated) condition as defined by
+%   ``:- if(Cond)`` under which the calling code is compiled.
+%   @arg Line is the _start line_ of the calling clause.
 
 xref_called(Source, Called, By) :-
     xref_called(Source, Called, By, _).
 
 xref_called(Source, Called, By, Cond) :-
     canonical_source(Source, Src),
-    called(Called, Src, By, Cond).
+    distinct(Called-By, called(Called, Src, By, Cond, _)).
 
+xref_called(Source, Called, By, Cond, Line) :-
+    canonical_source(Source, Src),
+    called(Called, Src, By, Cond, Line).
 
 %!  xref_defined(?Source, +Goal, ?How) is nondet.
 %
@@ -1011,19 +1031,19 @@ process_directive(arithmetic_function(FSpec), Src) :-
     arith_callable(FSpec, Goal),
     !,
     current_source_line(Line),
-    assert_called(Src, '<directive>'(Line), Goal).
+    assert_called(Src, '<directive>'(Line), Goal, Line).
 process_directive(format_predicate(_, Goal), Src) :-
     !,
     current_source_line(Line),
-    assert_called(Src, '<directive>'(Line), Goal).
+    assert_called(Src, '<directive>'(Line), Goal, Line).
 process_directive(if(Cond), Src) :-
     !,
     current_source_line(Line),
-    assert_called(Src, '<directive>'(Line), Cond).
+    assert_called(Src, '<directive>'(Line), Cond, Line).
 process_directive(elif(Cond), Src) :-
     !,
     current_source_line(Line),
-    assert_called(Src, '<directive>'(Line), Cond).
+    assert_called(Src, '<directive>'(Line), Cond, Line).
 process_directive(else, _) :- !.
 process_directive(endif, _) :- !.
 process_directive(Goal, Src) :-
@@ -1183,6 +1203,7 @@ xref_meta(phrase(G, _A, _R),    [//(G)]).
 xref_meta(call_dcg(G, _A, _R),  [//(G)]).
 xref_meta(phrase_from_file(G,_),[//(G)]).
 xref_meta(catch(A, _, B),       [A, B]).
+xref_meta(catch_with_backtrace(A, _, B), [A, B]).
 xref_meta(thread_create(A,_,_), [A]).
 xref_meta(thread_create(A,_),   [A]).
 xref_meta(thread_signal(_,A),   [A]).
@@ -1377,7 +1398,8 @@ process_goal(Goal, Origin, Src, P) :-
     ),
     !,
     must_be(list, Called),
-    assert_called(Src, Origin, Goal),
+    current_source_line(Here),
+    assert_called(Src, Origin, Goal, Here),
     process_called_list(Called, Origin, Src, P).
 process_goal(Goal, Origin, Src, _) :-
     process_xpce_goal(Goal, Origin, Src),
@@ -1393,16 +1415,19 @@ process_goal(use_foreign_library(File, _Init), _Origin, Src, _) :-
 process_goal(Goal, Origin, Src, P) :-
     xref_meta_src(Goal, Metas, Src),
     !,
-    assert_called(Src, Origin, Goal),
+    current_source_line(Here),
+    assert_called(Src, Origin, Goal, Here),
     process_called_list(Metas, Origin, Src, P).
 process_goal(Goal, Origin, Src, _) :-
     asserting_goal(Goal, Rule),
     !,
-    assert_called(Src, Origin, Goal),
+    current_source_line(Here),
+    assert_called(Src, Origin, Goal, Here),
     process_assert(Rule, Origin, Src).
 process_goal(Goal, Origin, Src, P) :-
     partial_evaluate(Goal, P),
-    assert_called(Src, Origin, Goal).
+    current_source_line(Here),
+    assert_called(Src, Origin, Goal, Here).
 
 disjunction(Var)   --> {var(Var), !}, [Var].
 disjunction((A;B)) --> !, disjunction(A), disjunction(B).
@@ -1602,7 +1627,8 @@ pce_goal(get_object(_,_,_), get_object(arg, msg, -)).
 process_xpce_goal(G, Origin, Src) :-
     pce_goal(G, Process),
     !,
-    assert_called(Src, Origin, G),
+    current_source_line(Here),
+    assert_called(Src, Origin, G, Here),
     (   arg(I, Process, How),
         arg(I, G, Term),
         process_xpce_arg(How, Term, Origin, Src),
@@ -1766,6 +1792,10 @@ process_use_module2(File, Import, Src, Reexport) :-
 %     * silent(+Boolean)
 %     Do not print any messages or raise exceptions on errors.
 %
+%   The information collected by this predicate   is  cached. The cached
+%   data is considered valid as long  as   the  modification time of the
+%   file does not change.
+%
 %   @param Source is the file from which Spec is referenced.
 
 xref_public_list(File, Src, Options) :-
@@ -1807,28 +1837,71 @@ xref_public_list(File, Path, Module, Export, Public, Meta, Src) :-
     xref_source_file(File, Path, Src),
     public_list(Path, Module, Meta, Export, Public, []).
 
-public_list(Path, Module, Meta, Export, Public, Options) :-
-    public_list_diff(Path, Module, Meta, [], Export, [], Public, [], Options).
+%!  public_list(+Path, -Module, -Meta, -Export, -Public, +Options)
+%
+%   Read the public information for Path.  Options supported are:
+%
+%     - silent(+Boolean)
+%       If `true`, ignore (syntax) errors.  If not specified the default
+%       is inherited from xref_source/2.
 
-public_list_diff(Path, Module, Meta, MT, Export, Rest, Public, PT, Options) :-
+:- dynamic  public_list_cache/6.
+:- volatile public_list_cache/6.
+
+public_list(Path, Module, Meta, Export, Public, _Options) :-
+    public_list_cache(Path, Modified,
+                      Module0, Meta0, Export0, Public0),
+    time_file(Path, ModifiedNow),
+    (   abs(Modified-ModifiedNow) < 0.0001
+    ->  !,
+        t(Module,Meta,Export,Public) = t(Module0,Meta0,Export0,Public0)
+    ;   retractall(public_list_cache(Path, _, _, _, _, _)),
+        fail
+    ).
+public_list(Path, Module, Meta, Export, Public, Options) :-
+    public_list_nc(Path, Module0, Meta0, Export0, Public0, Options),
+    (   Error = error(_,_),
+        catch(time_file(Path, Modified), Error, fail)
+    ->  asserta(public_list_cache(Path, Modified,
+                                  Module0, Meta0, Export0, Public0))
+    ;   true
+    ),
+    t(Module,Meta,Export,Public) = t(Module0,Meta0,Export0,Public0).
+
+public_list_nc(Path, Module, Meta, Export, Public, Options) :-
+    in_temporary_module(
+        TempModule,
+        true,
+        public_list_diff(TempModule, Path, Module,
+                         Meta, [], Export, [], Public, [], Options)).
+
+
+public_list_diff(TempModule,
+                 Path, Module, Meta, MT, Export, Rest, Public, PT, Options) :-
     setup_call_cleanup(
-        ( prolog_open_source(Path, In),
-          set_xref(Old)
-        ),
+        public_list_setup(TempModule, Path, In, State),
         phrase(read_directives(In, Options, [true]), Directives),
-        ( set_prolog_flag(xref, Old),
-          prolog_close_source(In)
-        )),
+        public_list_cleanup(In, State)),
     public_list(Directives, Path, Module, Meta, MT, Export, Rest, Public, PT).
 
+public_list_setup(TempModule, Path, In, state(OldM, OldXref)) :-
+    prolog_open_source(Path, In),
+    '$set_source_module'(OldM, TempModule),
+    set_xref(OldXref).
+
+public_list_cleanup(In, state(OldM, OldXref)) :-
+    '$set_source_module'(OldM),
+    set_prolog_flag(xref, OldXref),
+    prolog_close_source(In).
+
 
 read_directives(In, Options, State) -->
     {  repeat,
-         catch(prolog_read_source_term(In, Term, Expanded,
-                                       [ process_comment(true),
-                                         syntax_errors(error)
-                                       ]),
-               E, report_syntax_error(E, -, Options))
+       catch(prolog_read_source_term(In, Term, Expanded,
+                                     [ process_comment(true),
+                                       syntax_errors(error)
+                                     ]),
+             E, report_syntax_error(E, -, Options))
     -> nonvar(Term),
        Term = (:-_)
     },
@@ -1899,15 +1972,25 @@ public_list_1(meta_predicate(Decl), _Path, Meta, MT, Export, Export, Public, Pub
 public_list_1(public(Decl), _Path, Meta, Meta, Export, Export, Public, PT) :-
     phrase(public_decls(Decl), Public, PT).
 
+%!  reexport_files(+Files, +Src,
+%!                 -Meta, ?MetaTail, -Exports, ?ExportsTail,
+%!                 -Public, ?PublicTail)
+
 reexport_files([], _, Meta, Meta, Export, Export, Public, Public) :- !.
-reexport_files([H|T], Src, Meta, MT, Export, Rest, Public, PT) :-
+reexport_files([H|T], Src, Meta, MT, Export, ET, Public, PT) :-
     !,
     xref_source_file(H, Path, Src),
-    public_list_diff(Path, _, Meta, MT0, Export, Rest0, Public, PT0, []),
-    reexport_files(T, Src, MT0, MT, Rest0, Rest, PT0, PT).
-reexport_files(Spec, Src, Meta, MT, Export, Rest, Public, PT) :-
+    public_list(Path, _Module, Meta0, Export0, Public0, []),
+    append(Meta0, MT1, Meta),
+    append(Export0, ET1, Export),
+    append(Public0, PT1, Public),
+    reexport_files(T, Src, MT1, MT, ET1, ET, PT1, PT).
+reexport_files(Spec, Src, Meta, MT, Export, ET, Public, PT) :-
     xref_source_file(Spec, Path, Src),
-    public_list_diff(Path, _, Meta, MT, Export, Rest, Public, PT, []).
+    public_list(Path, _Module, Meta0, Export0, Public0, []),
+    append(Meta0, MT, Meta),
+    append(Export0, ET, Export),
+    append(Public0, PT, Public).
 
 public_from_import(except(Map), Path, Src, Export, Rest) :-
     !,
@@ -2154,53 +2237,53 @@ assert_constraint(Src, Head) :-
                 *       PHASE 1 ASSERTIONS      *
                 ********************************/
 
-%!  assert_called(+Src, +From, +Head) is det.
+%!  assert_called(+Src, +From, +Head, +Line) is det.
 %
 %   Assert the fact that Head is called by From in Src. We do not
 %   assert called system predicates.
 
-assert_called(_, _, Var) :-
+assert_called(_, _, Var, _) :-
     var(Var),
     !.
-assert_called(Src, From, Goal) :-
+assert_called(Src, From, Goal, Line) :-
     var(From),
     !,
-    assert_called(Src, '<unknown>', Goal).
-assert_called(_, _, Goal) :-
+    assert_called(Src, '<unknown>', Goal, Line).
+assert_called(_, _, Goal, _) :-
     expand_hide_called(Goal),
     !.
-assert_called(Src, Origin, M:G) :-
+assert_called(Src, Origin, M:G, Line) :-
     !,
     (   atom(M),
         callable(G)
     ->  current_condition(Cond),
         (   xmodule(M, Src)         % explicit call to own module
-        ->  assert_called(Src, Origin, G)
-        ;   called(M:G, Src, Origin, Cond) % already registered
+        ->  assert_called(Src, Origin, G, Line)
+        ;   called(M:G, Src, Origin, Cond, Line) % already registered
         ->  true
         ;   hide_called(M:G, Src)           % not interesting (now)
         ->  true
         ;   generalise(Origin, OTerm),
             generalise(G, GTerm)
-        ->  assert(called(M:GTerm, Src, OTerm, Cond))
+        ->  assert(called(M:GTerm, Src, OTerm, Cond, Line))
         ;   true
         )
     ;   true                        % call to variable module
     ).
-assert_called(Src, _, Goal) :-
+assert_called(Src, _, Goal, _) :-
     (   xmodule(M, Src)
     ->  M \== system
     ;   M = user
     ),
     hide_called(M:Goal, Src),
     !.
-assert_called(Src, Origin, Goal) :-
+assert_called(Src, Origin, Goal, Line) :-
     current_condition(Cond),
-    (   called(Goal, Src, Origin, Cond)
+    (   called(Goal, Src, Origin, Cond, Line)
     ->  true
     ;   generalise(Origin, OTerm),
         generalise(Goal, Term)
-    ->  assert(called(Term, Src, OTerm, Cond))
+    ->  assert(called(Term, Src, OTerm, Cond, Line))
     ;   true
     ).
 
@@ -2260,7 +2343,7 @@ assert_import(Src, Import as Name, Export, From, Reexport) :-
     ->  assert(imported(Term, Src, From)),
         assert_reexport(Reexport, Src, Term)
     ;   current_source_line(Line),
-        assert_called(Src, '<directive>'(Line), Term0)
+        assert_called(Src, '<directive>'(Line), Term0, Line)
     ).
 assert_import(Src, Import, Export, From, Reexport) :-
     pi_to_head(Import, Term),
@@ -2269,7 +2352,7 @@ assert_import(Src, Import, Export, From, Reexport) :-
     ->  assert(imported(Term, Src, From)),
         assert_reexport(Reexport, Src, Term)
     ;   current_source_line(Line),
-        assert_called(Src, '<directive>'(Line), Term)
+        assert_called(Src, '<directive>'(Line), Term, Line)
     ).
 assert_import(Src, op(P,T,N), _, _, _) :-
     xref_push_op(Src, P,T,N).
@@ -2430,7 +2513,7 @@ assert_multifile(PI, Src) :-                    % :- multifile(Spec)
 assert_public(PI, Src) :-                       % :- public(Spec)
     pi_to_head(PI, Term),
     current_source_line(Line),
-    assert_called(Src, '<public>'(Line), Term),
+    assert_called(Src, '<public>'(Line), Term, Line),
     assert(public(Term, Src, Line)).
 
 assert_export(PI, Src) :-                       % :- export(Spec)
