@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2016, VU University Amsterdam
+    Copyright (c)  2013-2019, VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -432,6 +433,7 @@ verify_safe_declaration(Var) :-
     !,
     instantiation_error(Var).
 verify_safe_declaration(Module:Goal) :-
+    !,
     must_be(atom, Module),
     must_be(callable, Goal),
     (   ok_meta(Module:Goal)
@@ -538,6 +540,8 @@ safe_primitive(=:=(_,_)).
 safe_primitive(=\=(_,_)).
 safe_primitive(=<(_,_)).
 safe_primitive(<(_,_)).
+safe_primitive(system:nth_integer_root_and_remainder(_,_,_,_)).
+
                                         % term-handling
 safe_primitive(arg(_,_,_)).
 safe_primitive(system:setarg(_,_,_)).
@@ -702,6 +706,8 @@ safe_primitive(system:b_setval(Var,_)) :-
 safe_primitive(system:nb_getval(_,_)).
 safe_primitive('$syspreds':nb_setval(Var,_)) :-
     safe_global_var(Var).
+safe_primitive(system:nb_linkval(Var,_)) :-
+    safe_global_var(Var).
 safe_primitive(system:nb_current(_,_)).
                                         % database
 safe_primitive(system:assert(X)) :-
@@ -839,6 +845,7 @@ safe_meta('$tabling':abolish_table_subgoals(V), []) :-
     \+ qualified(V).
 safe_meta('$tabling':current_table(V, _), []) :-
     \+ qualified(V).
+safe_meta('$tabling':tnot(G), [G]).
 
 qualified(V) :-
     nonvar(V),
@@ -1006,6 +1013,7 @@ safe_meta(call(4,*,*,*,*)).
 safe_meta(call(5,*,*,*,*,*)).
 safe_meta(call(6,*,*,*,*,*,*)).
 safe_meta('$tabling':start_tabling(*,0)).
+safe_meta('$tabling':start_tabling(*,0,*,*)).
 
 %!  safe_output(+Output)
 %
